//
//  MSchemaDataSource.h
//  MySQLGUICommon
//
//  Created by Alfredo Kojima on Sun Jun 27 2004.
//  Copyright (c) 2004 MySQL AB. All rights reserved.
//

#import <Foundation/Foundation.h>
#include "myx_public_interface.h"


typedef enum {
  MCatalogItemType,
  MSchemaItemType,
  MTableItemType,
  MColumnItemType
} MSchemaDSItemType;

@class NSImage;

@interface MSchemaItem : NSObject {
@public
  MSchemaDSItemType type;
  NSImage *icon;
  MSchemaItem *parent;
@protected
  union {
    MYX_CATALOG *catalog;
    MYX_SCHEMA *schema;
    MYX_SCHEMA_TABLE *table;
    MYX_SCHEMA_TABLE_COLUMN *column;
  } _object;
  NSString *_repr;
  id _identifier;
}

- (id)initWithCatalog: (MYX_CATALOG*)catalog
                 icon: (NSImage*)icon;
- (id)initWithSchema: (MYX_SCHEMA*)schema
	   parentCatalog: (MSchemaItem*)parent
                icon: (NSImage*)icon;
- (id)initWithTable: (MYX_SCHEMA_TABLE*)table
	   parentSchema: (MSchemaItem*)parent
               icon: (NSImage*)icon;
- (id)initWithColumn: (MYX_SCHEMA_TABLE_COLUMN*)column
		 parentTable: (MSchemaItem*)parent
                icon: (NSImage*)icon;

- (void)dealloc;

- (id)identifier;

- (NSString*)repr;
- (NSImage*)icon;

- (MYX_CATALOG*)catalog;
- (MYX_SCHEMA*)schema;
- (MYX_SCHEMA_TABLE*)table;
- (MYX_SCHEMA_TABLE_COLUMN*)column;

@end


@interface MSchemaDataSource : NSObject {
@public
  MSchemaDSItemType _rootType;
  MSchemaDSItemType _leafType;

  id _tableFetcher;
  SEL _tableFetcherAction;

  NSMutableDictionary *_children; // maps identifier -> NSArray of identifiers
  NSMutableDictionary *_items; // maps identifier -> MSchemaItem
}

- (id)initWithRoot: (MSchemaDSItemType)root
              leaf: (MSchemaDSItemType)leaf;
- (void)dealloc;

- (void)resetTree;
- (void)setChildren: (NSMutableArray*)array
			forItem: (MSchemaItem*)item;

- (void)setTableFetcher: (id)target selector:(SEL)sel;

- (MSchemaItem*)findItem: (NSString*)name;

- (id)outlineView:(NSOutlineView *)outlineView 
            child:(int)index 
           ofItem:(id)item;

- (BOOL)outlineView:(NSOutlineView *)outlineView 
   isItemExpandable:(id)item;

- (int)outlineView:(NSOutlineView *)outlineView 
numberOfChildrenOfItem:(id)item;

- (id)outlineView:(NSOutlineView *)outlineView 
objectValueForTableColumn:(NSTableColumn *)tableColumn 
           byItem:(id)item;

@end



@interface MFilteredSchemaDataSource : MSchemaDataSource {
  NSString *_filter;
  MSchemaDataSource *_parentDS;
}

- (id)initWithDataSource:(MSchemaDataSource*)parent;
- (void)performSearch:(NSString*)filter;


@end






